use std::time::Duration;
use std::{thread, time};
use serialport::{SerialPort, StopBits, Parity, FlowControl, DataBits};
use std::io::{self, Read, Write, BufWriter};
use regex::Regex;
use rustyline::Editor;
use rustyline::error::ReadlineError;
use std::fs::OpenOptions;
use std::io::{BufReader, BufRead};
use std::net::{TcpListener, TcpStream};
use std::collections::HashMap;

pub mod perf {
    use chrono::{DateTime, Duration, Utc, Local};

    pub struct Perf {
        start: DateTime<Utc>,
        end: DateTime<Utc>,
        delta_ms: i64,
        delta_mks: i64,
        delta_ns: i64,
    }

    impl Perf  {
        pub fn new() -> Perf {
            Perf {
                start: Utc::now(),
                end: Utc::now(),
                delta_ms: 0,
                delta_mks: 0,
                delta_ns: 0,
            }
        }

        pub fn start(&mut self) {
            self.start = Utc::now();
        }

        // measure: "ms" or "mks"
        pub fn stop(&mut self, out_msg: &str, measure: &str) {
            self.end = Utc::now();
            self.delta_ms = (self.end - self.start).num_microseconds().unwrap()/1000;
            self.delta_mks = (self.end - self.start).num_microseconds().unwrap();
            self.delta_ns = (self.end - self.start).num_nanoseconds().unwrap();

            if (measure == "mks") { println!("{} (mks): {:?}", out_msg, self.delta_mks); }
            else if (measure == "ns") { println!("{} (ns): {:?}", out_msg, self.delta_ns); }
            else { println!("{} (ms): {:?}", out_msg, self.delta_ms); }
        }

        pub fn stop_restart(&mut self, out_msg: &str, measure: &str) {
            self.stop(out_msg, measure);
            self.start = Utc::now();
        }

        pub fn cur_time(format: &str) -> String {
            let now = Local::now();
            let fmt = now.format(format).to_string();
            return fmt;
        }

    }
}

use crate::perf::Perf;

fn tcp_server() {
    let listener = TcpListener::bind("127.0.0.1:10000").unwrap();
    println!("TCP-server is running");

    for stream in listener.incoming() {
        match stream {
            Ok(stream) => {
                println!("connect to client: {:?}", stream);
                handle_client(stream);

            },
            Err(ref msg) => {
                println!("error: failed attempt to set connection with client: ({:?}): {}", 
                         stream, msg);
            }
        }
    }
}

fn handle_client(mut stream: TcpStream) {
    let mut cmd_descr: HashMap::<&str, &str> = HashMap::new();
    cmd_descr.insert("send", "send data to serial port");
    cmd_descr.insert("exit", "cancel connection with server");

    loop {
        let mut reader = BufReader::new(&stream);
        let mut rx_data = String::new();
        if let Err(msg) = reader.read_line(&mut rx_data) {
            println!("error: failed attempt to read data from client: {}", msg);
            return;
        }

        if rx_data.contains("exit") {
            println!("The client requested to close the connection");
            return;
        }
    }
}


pub mod uartTransact {
    const MAX_BUF: usize = 4096;
    use serialport::{SerialPort, StopBits, Parity, FlowControl, DataBits};
    use std::time::Duration;
    use crate::perf::Perf;
    use std::io::{self, Read, Write, BufWriter, BufReader, BufRead};
    use std::fs::{File, OpenOptions};
    use std::path::Path;

    #[derive(PartialEq)]
    pub enum Type_en {
        rx = 0,
        tx,
        rx_tx,
        tx_rx,
    }

    #[derive(Clone)]
    pub struct PortSettings_t {
        baud_rate: u32,
        data_bits: DataBits,
        parity: Parity,
        stop_bits: StopBits,
        flow_ctrl: FlowControl,
        timeout_msec: u64,
    }

    impl PortSettings_t {
        pub fn new(baud_rate: u32, data_bits_: u8, parity_: &str, stop_bits_: u8,
                   flow_ctrl_: &str, timeout_msec: u64) -> Result<Self, String> {

            let data_bits = match data_bits_ {
                5 => DataBits::Five,
                6 => DataBits::Six,
                7 => DataBits::Seven,
                8 => DataBits::Eight,
                _ => { return Err(format!(
                            "error: the number of data bits specified is incorrect 
                            (type from 5 to 8)")); 
                }
            };

            let parity = match parity_ {
                "Even" => Parity::Even,
                "Odd" => Parity::Odd,
                "None" => Parity::None,
                _ => { return Err(format!(
                            "error: the parity of data specified is incorrect 
                            (type \"Even\", \"Odd\" or \"None\")")); 
                }
            };

            let stop_bits = match stop_bits_ {
                1 => StopBits::One,
                2 => StopBits::Two,
                _ => { return Err(format!(
                            "error: the stop bits of data specified is incorrect 
                            (type 1 or 2)")); 
                }
            };

            let flow_ctrl = match flow_ctrl_ {
                "None" => FlowControl::None,
                "Software" => FlowControl::Software,
                "Hardware" => FlowControl::Hardware,
                _ => { return Err(format!(
                            "error: the parity of data specified is incorrect 
                            (type \"None\", \"Software\" or \"Hardware\")")); 
                }
            };

            Ok(Self { baud_rate, data_bits, parity, stop_bits, flow_ctrl, timeout_msec })
        }
    }

    struct Packet_t {
        buf: [u8; MAX_BUF],
        size: u32,
    }

    impl Packet_t {
        pub fn new(buf: [u8; MAX_BUF], size: u32) -> Packet_t {
            Packet_t { buf, size }
        }
    }

    struct Logger {
        out: BufWriter<File>, 
        log_flag: bool,
    }

    impl Logger {
        fn new<P: AsRef<Path>>(filename: P, log_flag: bool) -> Result<Self, io::Error> {
            let path = Path::new(filename.as_ref());

            let mut file = match path.exists() {
                true => OpenOptions::new().write(true).append(true).open(filename.as_ref())?,
                false => File::create(filename.as_ref())?
            };

            let out = BufWriter::new(file);

            Ok(Self { out, log_flag })
        }

        fn log(&mut self, msg: &str) -> Result<(), io::Error> {
            if self.log_flag == true {
                writeln!(&mut self.out, "{} {};{:?}", 
                         Perf::cur_time("%Y.%m.%d"), 
                         Perf::cur_time("%H:%M:%S.%3f"),
                         msg.trim())?;
                self.out.flush()?;
            }
            Ok(())
        }
    }

    pub struct UartTransact {
        type_: Type_en,
        port_name: String,
        port_settings: PortSettings_t,
        tx_packet: String,
        rx_packet: String,

        port: Box<dyn SerialPort>, 
        save_data_flag: bool,
        save_cmd_flag: bool,
        data_writer: Logger,
        cmd_writer: Logger,
    }

    impl UartTransact {
        pub fn new(trans_type: &str, 
                   port_name: &str, 
                   port_settings: PortSettings_t,
                   save_data_flag: bool,
                   save_cmd_flag: bool) -> Result<Self, io::Error> {

            let mut type_ = Type_en::tx_rx;
            if trans_type == "tx_rx" { type_ = Type_en::tx_rx;  }
            else if trans_type == "tx" { type_ = Type_en::rx;  }
            else if trans_type == "rx_tx" { type_ = Type_en::rx_tx;  }
            else if trans_type == "rx" { type_ = Type_en::rx;  }

            let mut filename = format!("../log/data_{}.log", Perf::cur_time("%Y%m%d")); 
            let data_writer = Logger::new(filename.clone(), save_data_flag)?; 
            filename = format!("../log/commands_{}.log", Perf::cur_time("%Y%m%d")); 
            let cmd_writer = Logger::new(filename.clone(), save_data_flag)?; 

            let mut port = serialport::new(port_name.clone(), port_settings.baud_rate)
                .timeout(Duration::from_millis(port_settings.timeout_msec)).open().unwrap();

            let tx_packet = String::new();
            let rx_packet = String::new();

            Ok( Self { type_, port_name: port_name.to_string(), port_settings, save_data_flag, save_cmd_flag, 
                data_writer, cmd_writer,
                port, tx_packet, rx_packet,
            })
        }


        pub fn addPacketForTx(&mut self, buf: String, size: u32) -> Result<(), String> {
            if size as usize >= MAX_BUF {
                return Err(format!("too much data for transmit! Data size must be < {} bytes", MAX_BUF));
            }
            else { self.tx_packet = buf; }

            return Ok(());
        }

        pub fn startTransact(&mut self) -> Result<(), String> {

            if self.type_ == Type_en::tx_rx  {
                self.cmd_writer.log(&self.tx_packet).unwrap();

                match self.port.write(&self.tx_packet.as_bytes()) {
                    Ok(_) => {},
                    Err(e) => eprintln!("Error sending data: {}", e),
                }

                // self.port.flush().unwrap();

                let mut reader = BufReader::new(&mut self.port);

                let mut s = String::new();
                if let Err(err) = reader.read_line(&mut s) {
                    println!("error: {}", err);
                    return Ok(());
                }
                if s.len() > 0 {
                    self.rx_packet = s.clone();
                    self.data_writer.log(&s).unwrap();
                }
            }
            Ok(())
        }

        pub fn getReceivedPacket(&mut self) -> String {
            self.rx_packet.clone()
        }

        fn set_port_settings(port_name: &str, 
                             set: PortSettings_t) -> Result<Box<dyn SerialPort>, String> {

            let mut port = match serialport::new(port_name, set.baud_rate)
                .timeout(Duration::from_millis(set.timeout_msec)).open() {
                    Ok(p) => p,
                    Err(msg) => return Err(format!("error: failed to open serial port {} ({})", port_name, msg)),
                };

            match port.set_baud_rate(set.baud_rate) {
                Ok(_) => {},
                Err(msg) => return Err(format!("error: failed to open serial port {} ({})", port_name, msg)),
            }
            match port.set_data_bits(set.data_bits) {
                Ok(_) => {},
                Err(msg) => return Err(format!("error: failed to open serial port {} ({})", port_name, msg)),
            }
            match port.set_parity(set.parity) {
                Ok(_) => {},
                Err(msg) => return Err(format!("error: failed to open serial port {} ({})", port_name, msg)),
            }
            match port.set_stop_bits(set.stop_bits) {
                Ok(_) => {},
                Err(msg) => return Err(format!("error: failed to open serial port {} ({})", port_name, msg)),
            }
            match port.set_flow_control(set.flow_ctrl) {
                Ok(_) => {},
                Err(msg) => return Err(format!("error: failed to open serial port {} ({})", port_name, msg)),
            }

            return Ok(port);
        }
    }
}

fn read_str(s: &str) -> Result<Vec<String>, String> { 
    let re = match Regex::new(r#""([^"]+)"|(\S+)"#) {
        Ok(val) => val,
        Err(_) => { return Err(format!("Error in regex!"));  }
    };

    let res: Vec<String> = re.find_iter(&s)
        .map(|match_| match_.as_str().to_owned().replace("\"", ""))
        .collect();

    Ok(res)
}





fn main() {
    use crate::perf::Perf;
    use crate::uartTransact::*;
    let mut perf = Perf::new();

    let fname_hist = "../settings/cmd_history.log".to_string();
    let mut editor = Editor::<()>::new();
    if editor.load_history(&fname_hist).is_err() {
        println!("error while reading cmd_history.log file!", ); 
        return;
    } 

    loop {
        let s: String = match editor.readline(">> ") {
            Ok(s_) => s_.to_string(),
            Err(ReadlineError::Interrupted) => {
                editor.save_history(&fname_hist)
                    .expect("error while save the command to the history file (cmd_history.log)"); 
                return; 
            },
            Err(ReadlineError::Eof) => {
                editor.save_history(&fname_hist) 
                    .expect("error while save the command to the history file (cmd_history.log)"); 
                return; 
            },
            Err(msg) => {
                println!("{}{:?}",
                         "unknown command",
                         msg);
                return;
            }
        };

        editor.add_history_entry(&s);


        let args = match read_str(&s) {
            Ok(arg) => arg,
            Err(msg) => { 
                println!("{}", msg);
                continue;
            }
        };

        let args_cnt = args.len();

        let cmd_head = &args[0].clone();

        if cmd_head.contains("exit") { break; }
        else if cmd_head.contains("help") {
            println!("Serial port terminal supports the following commands:");
            println!("\nexit - excecute quit from this program");
            println!("\nhelp - explain how to use this program");
            println!("\nshow_ports - shows all available serial ports");
            println!("\nsend <port_name> <data> - sends specified <data>");
            println!("\nsend_loop <port name> <data> <count times>  <period in msec> \n\tsends specified <data> to <port name> in loop mode: <count times> times with period specified in <period in msec> ");
            println!("\nopen_port <port name> <speed> <data bits: 5, 6, 7, 8> <parity: None, Even, Odd> <stop bits: 1, 2> <flow control: None, Software, Hardware> <timeout_ms> <transaction type: tx, rx, tx_rx, rx_tx> <save commands and data to file: true, false>.\n\tExample1: open_port /dev/ttyUSB0 9600 8 Even 1 None 1000 tx_rx true");
        }
        else if cmd_head.contains("show_ports") {
            match serialport::available_ports() {
                Ok(ports) => {
                    for port in ports {
                        println!("{}: {:?}", port.port_name, port.port_type);
                    }
                }
                Err(msg) => { eprintln!("error while get all available serial port list: {}", msg); return; }
            }
        }
        else if cmd_head == "send" {
            if args_cnt != 3 { eprintln!("Error while parse command! {}", "arguments count must be 2"); }
            else {
                let port_name = &args[1];
                let timeout: u64 = 200;
                let port_set = PortSettings_t::new(9600, 8, "None", 1, "None", timeout).unwrap();
                let save_data_flag_ = true;
                let save_cmd_flag_ = true;
                let trans_type = "tx_rx";

                let mut uartTrans = UartTransact::new(trans_type, port_name, 
                                                      port_set, save_data_flag_, save_cmd_flag_).unwrap(); 

                let mut perf = Perf::new();
                perf.start();
                let mut data_out = args[2].clone();
                data_out.push_str(std::str::from_utf8(b"\x0D\x0A").unwrap()); // CR + LF

                uartTrans.addPacketForTx(data_out.clone(), data_out.len() as u32).unwrap();
                uartTrans.startTransact().unwrap();

                let s = uartTrans.getReceivedPacket();
                let data = format!("{}: {}", Perf::cur_time("%H:%M:%S.%f"), s.trim());
                println!("{}", data);
                perf.stop("time for transaction", "mks");
            }
        }
        else if cmd_head == "send_loop" {
            if args_cnt != 5 { eprintln!("Error while parse command! {}", "arguments count must be 4"); }
            else {
                let cnt = args[3].parse::<u32>().unwrap();
                let period = args[4].parse::<u64>().unwrap();
                let mut i = 0;
                while i < cnt {
                    i += 1;
                    let mut data_out = args[2].clone();
                    thread::sleep(time::Duration::from_millis(period));
                }
            }
        }
        else if cmd_head.contains("open_port") {

            if args_cnt != 10 { eprintln!("Error while parse command! {}", "arguments count must be 9"); }
            else {
                let port_name = &args[1];
                let speed = args[3].parse::<u32>().unwrap();
                let data_bits = args[3].parse::<u8>().unwrap();
                let parity = &args[4];
                let stop_bits = args[5].parse::<u8>().unwrap();
                let flow_ctrl = &args[6];
                let timeout = args[7].parse::<u64>().unwrap();
                let trans_type = &args[8];
                let save_data_flag = args[9].parse::<bool>().unwrap(); 
                let save_cmd_flag = save_data_flag; 

                let port_set = PortSettings_t::new(speed, data_bits, parity, stop_bits, 
                                                   flow_ctrl, timeout).unwrap();

                let mut uartTrans = UartTransact::new(trans_type, &port_name, 
                                              port_set, save_data_flag, save_cmd_flag).unwrap(); 
            }
        }
    }

    // close ports
    editor.save_history(&fname_hist).unwrap();
    // drop(port1);

}

fn print_type_of<T>(_: &T) {
    println!("{}", std::any::type_name::<T>());
}


// perf.stop("time for transaction", "mks");

// let mut buffer = Vec::new();
// match port.read_to_end(&mut buffer) {
//     Ok(_) => {
//         if buffer.len() > 0 {
//             let data = format!("{}: {:?}", cur_time("%H:%M:%S"), buffer);
//             println!("{}", data);
//             if save_data_flag == true {
//                 writeln!(data_writer, "{}", 
//                          format!("{};{};{}", 
//                          cur_time("%Y.%m.%d"), 
//                          cur_time("%H:%M:%S"), 
//                          String::from_utf8(buffer).expect("msg")))
//                     .expect("err");
//             }
//         }
//     },
//     Err(msg) => eprintln!("{}", msg),
// }

// let mut cnt: u32 = 0;
// let mut i = 0;
// let mut buf: String = String::new();

// let start = Utc::now(); 
// let mut timeout_flag = false;

// let mut delta1 = 0;
// while cnt == 0 {
//     cnt = match port.bytes_to_read() {
//         Ok(c) => c,
//         Err(msg) => { println!("{}", msg); return; }
//     };
//     let delta = Utc::now() - start;
//     delta1 =  delta.num_microseconds().unwrap()/1000;
//     if delta1 > timeout_ms {
//         println!("timeout: {}", delta1);
//         timeout_flag = true;
//         break;
//     };
// }

// println!("time (ms): {}", delta1);

// // println!("time (ms): {:?}", delta.num_microseconds().unwrap()/1000);

// if timeout_flag == false {
//     while i < 500000 {
//         // thread::sleep(time::Duration::from_millis(10));
//         i += 1;
//         let mut buffer = [0u8; 4096];

//         cnt = match port.bytes_to_read() {
//             Ok(c) => c,
//             Err(msg) => { println!("{}", msg); return; }
//         };

//         if cnt > 0 {
//             match port.read(&mut buffer) {
//                 Ok(count) => {
//                     let s = std::str::from_utf8(&buffer[..count]).unwrap();
//                     buf.push_str(s);
//                     println!("{}", s);
//                 },

//                 Err(ref e) if e.kind() == io::ErrorKind::TimedOut => { println!("timeout!"); },
//                 Err(e) => eprintln!("error reading from port ({}): {}", port_name, e),
//             }
//         }
//         else {}
//     }

// }


// if buf.len() > 0 {
//     let data = format!("{}: {:?}", cur_time("%H:%M:%S"), buf);
//     println!("{}", data);
//     if save_data_flag == true {
//         writeln!(data_writer, "{};{};{}", cur_time("%Y.%m.%d"), cur_time("%H:%M:%S"), buf.trim()).expect("err");
//     }
// }

